import 'dart:io' as io;

import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/file_system/file_system.dart';
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:analyzer/source/package_map_provider.dart';
import 'package:analyzer/source/package_map_resolver.dart';
import 'package:analyzer/source/pub_package_map_provider.dart';
import 'package:analyzer/src/context/cache.dart';
import 'package:analyzer/src/context/context.dart';
import 'package:analyzer/src/generated/engine.dart';
import 'package:analyzer/src/generated/resolver.dart';
import 'package:analyzer/src/generated/sdk.dart';
import 'package:analyzer/src/generated/sdk_io.dart';
import 'package:analyzer/src/generated/source.dart';
import 'package:analyzer/src/generated/source_io.dart';
import 'package:analyzer/src/generated/utilities_general.dart';
import 'package:analyzer/src/summary/format.dart';
import 'package:analyzer/src/summary/resynthesize.dart';
import 'package:analyzer/src/summary/summarize_elements.dart';
import 'package:analyzer/src/summary/summary_sdk.dart';
import 'package:analyzer/src/task/dart.dart';
import 'package:analyzer/task/dart.dart';
import 'package:analyzer/task/model.dart';
import 'package:crypto/crypto.dart' as crypto;
import 'package:path/path.dart' as pathos;
import 'package:analyzer/src/summary/idl.dart';
import 'package:analyzer/src/generated/java_io.dart';

main() {
  main_onlySDK();
//  main_SDK_languageTests();
//  {
//    LibraryElement library = main_withPackages(true);
////    writeLibrarySummaries(new Set<LibraryElement>(), library);
//    print('----- Performance tags -----');
//    PerformanceTag.all
//        .forEach((tag) => print('${tag.label}: ${tag.elapsedMs}'));
//    print('----- Analysis Task -----');
//    AnalysisTask.stopwatchMap
//        .forEach((type, sw) => print('$type: ${sw.elapsedMilliseconds}'));
//  }
//  main_inCycle_parse();
//  main_inCycle_analyze(false);
}

void main_inCycle_analyze(bool useSummaries) {
  for (int i = 0; i < 30; i++) {
    AnalysisEngine.instance.clearCaches();
    Stopwatch sw = new Stopwatch()..start();
    {
      ResourceProvider resourceProvider = PhysicalResourceProvider.INSTANCE;
      DartSdk sdk = DirectoryBasedDartSdk.defaultSdk;
      PackageMapInfo packageMapInfo =
          new PubPackageMapProvider(resourceProvider, sdk).computePackageMap(
              resourceProvider.getFolder(
                  '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer'));
//      print('{');
//      packageMapInfo.packageMap.forEach((k, v) => print('\t$k: $v'));
//      print('}');

      AnalysisContextImpl context =
          AnalysisEngine.instance.createAnalysisContext();
      context.sourceFactory = new SourceFactory(<UriResolver>[
        new FileUriResolver(),
        new DartUriResolver(sdk),
        new PackageMapUriResolver(resourceProvider, packageMapInfo.packageMap)
      ]);

      if (useSummaries) {
//    context.typeProvider =
//        context.computeResult(AnalysisContextTarget.request, TYPE_PROVIDER);
        PerformanceStatistics.summary.makeCurrentWhile(() {
          context.resultProvider = new MySummaryResultProvider(context);
        });
      }

//  String path =
//      '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/lib/src/codegen/tools.dart';
      String path =
          '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/lib/file_system/file_system.dart';
//  String path =
//      '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/bin/analyze.dart';
      File file = resourceProvider.getFile(path);
      Source source = file.createSource();
      context.applyChanges(new ChangeSet()..addedSource(source));
      //
      context.computeLibraryElement(source);
    }
    print('$i: ${sw.elapsedMilliseconds}');
  }
}

main_inCycle_parse() {
  ResourceProvider resourceProvider = PhysicalResourceProvider.INSTANCE;
  String path =
      '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/lib/src/codegen/tools.dart';
  File file = resourceProvider.getFile(path);
  Source source = file.createSource();

  for (int i = 0; i < 100; i++) {
    var context = AnalysisEngine.instance.createAnalysisContext();
    DartSdk sdk = DirectoryBasedDartSdk.defaultSdk;
    context.sourceFactory = new SourceFactory(
        <UriResolver>[new FileUriResolver(), new DartUriResolver(sdk)]);
    Stopwatch sw = new Stopwatch()..start();
    context.parseCompilationUnit(source);
    print('$i: ${sw.elapsedMilliseconds}');
  }
}

main_onlySDK() {
  for (int i = 0; i < 1; i++) {
    Stopwatch sw = new Stopwatch()..start();
    AnalysisEngine.instance.clearCaches();
    var context = AnalysisEngine.instance.createAnalysisContext();
    DirectoryBasedDartSdk sdk = DirectoryBasedDartSdk.defaultSdk;
    sdk.useSummary = true;
    context.sourceFactory = new SourceFactory(
        <UriResolver>[new FileUriResolver(), new DartUriResolver(sdk)]);
    //
    ResourceProvider resourceProvider = PhysicalResourceProvider.INSTANCE;
    String path = '/Users/scheglov/dart/test/bin/test.dart';
//    String path = '/Users/scheglov/dart/test/bin/test2.dart';
//    String path = '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/lib/src/cancelable_future.dart';
//    String path =
//        '/Users/scheglov/Source/Dart/sdk.git/sdk/xcodebuild/ReleaseIA32/generated_tests/language/async_await_test_02.dart';
    File file = resourceProvider.getFile(path);
    Source source = file.createSource();
    context.applyChanges(new ChangeSet()..addedSource(source));
    //
    while (true) {
      var analysisResult = context.performAnalysisTask();
      if (!analysisResult.hasMoreWork) {
        break;
      }
    }
    print(context.getErrors(source).errors.join('\n'));
    print('$i: ${sw.elapsedMilliseconds}');
  }
}

void main_SDK_languageTests() {
  // TODO
  ResourceProvider resourceProvider = PhysicalResourceProvider.INSTANCE;
  DartSdk sdk = DirectoryBasedDartSdk.defaultSdk;
  Folder packageFolder = resourceProvider.getFolder(
      '/Users/scheglov/Source/Dart/sdk.git/sdk/xcodebuild/ReleaseIA32/packages');
  Map<String, List<Folder>> packageMap = <String, List<Folder>>{};
  packageFolder.getChildren().forEach((r) {
    if (r.exists && r is Folder) {
      packageMap[r.shortName] = <Folder>[r];
    }
  });
//  PackageMapInfo packageMapInfo =
//      new PubPackageMapProvider(resourceProvider, sdk)
//          .computePackageMap(packageFolder);
  print('{');
  packageMap.forEach((k, v) => print('\t$k: $v'));
  print('}');

  print(sdk.context);

  AnalysisContextImpl context = AnalysisEngine.instance.createAnalysisContext();
  context.sourceFactory = new SourceFactory(<UriResolver>[
    new FileUriResolver(),
    new DartUriResolver(sdk),
    new PackageUriResolver([new JavaFile(packageFolder.path)])
  ]);
//  context.sourceFactory = new SourceFactory(<UriResolver>[
//    new FileUriResolver(),
//    new DartUriResolver(sdk),
//    new PackageMapUriResolver(resourceProvider, packageMap)
//  ]);

//  if (useSummaries) {
//    PerformanceStatistics.summary.makeCurrentWhile(() {
//      context.resultProvider = new MySummaryResultProvider(context);
//    });
//  }

  String path =
      '/Users/scheglov/Source/Dart/sdk.git/sdk/xcodebuild/ReleaseIA32/generated_tests/language/assignable_expression_test_03.dart';
  File file = resourceProvider.getFile(path);
  Source source = file.createSource();
  context.applyChanges(new ChangeSet()..addedSource(source));
  //
  LibraryElement libraryElement = context.computeLibraryElement(source);
  print(context.getErrors(source).errors.join('\n'));
}

LibraryElement main_withPackages(bool useSummaries) {
  ResourceProvider resourceProvider = PhysicalResourceProvider.INSTANCE;
  DartSdk sdk = DirectoryBasedDartSdk.defaultSdk;
  PackageMapInfo packageMapInfo =
      new PubPackageMapProvider(resourceProvider, sdk).computePackageMap(
          resourceProvider.getFolder(
              '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer'));
  print('{');
  packageMapInfo.packageMap.forEach((k, v) => print('\t$k: $v'));
  print('}');

  AnalysisContextImpl context = AnalysisEngine.instance.createAnalysisContext();
  context.sourceFactory = new SourceFactory(<UriResolver>[
    new FileUriResolver(),
    new DartUriResolver(sdk),
    new PackageMapUriResolver(resourceProvider, packageMapInfo.packageMap)
  ]);

  if (useSummaries) {
//    context.typeProvider =
//        context.computeResult(AnalysisContextTarget.request, TYPE_PROVIDER);
    PerformanceStatistics.summary.makeCurrentWhile(() {
      context.resultProvider = new MySummaryResultProvider(context);
    });
  }

//  String path =
//      '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/lib/src/codegen/tools.dart';
  String path =
      '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/lib/file_system/file_system.dart';
//  String path =
//      '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/bin/analyze.dart';
  File file = resourceProvider.getFile(path);
  Source source = file.createSource();
  context.applyChanges(new ChangeSet()..addedSource(source));
  //
  LibraryElement libraryElement = context.computeLibraryElement(source);
  print(context.getErrors(source).errors.join('\n'));
  return libraryElement;
}

void writeLibrarySummaries(
    Set<LibraryElement> seenLibs, LibraryElement library) {
  if (seenLibs.add(library)) {
    Source source = library.source;
    String sumPath = _getSummaryPath(source.fullName);
    // TODO
    if (!source.isInSystemLibrary &&
        !source.fullName.startsWith(
            '/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer/lib')) {
      print('$library in $sumPath');
      List<String> prelinkedLibraryUris = <String>[];
      List<LinkedLibraryBuilder> prelinkedLibraries = <LinkedLibraryBuilder>[];
      List<String> unlinkedUnitUris = <String>[];
      List<UnlinkedUnitBuilder> unlinkedUnits = <UnlinkedUnitBuilder>[];
      {
        LibrarySerializationResult libraryResult =
            serializeLibrary(library, library.context.typeProvider, false);
        prelinkedLibraryUris.add(library.source.uri.toString());
        prelinkedLibraries.add(libraryResult.linked);
        unlinkedUnitUris.addAll(libraryResult.unitUris);
        unlinkedUnits.addAll(libraryResult.unlinkedUnits);
      }
      //
      // Write the whole bundle.
      //
//      print('  prelinkedLibraryUris: $prelinkedLibraryUris');
//      print('  unlinkedUnitUris: $unlinkedUnitUris');
      PackageBundleBuilder bundle = new PackageBundleBuilder(
          linkedLibraryUris: prelinkedLibraryUris,
          linkedLibraries: prelinkedLibraries,
          unlinkedUnitUris: unlinkedUnitUris,
          unlinkedUnits: unlinkedUnits);
      io.File file = new io.File(sumPath);
      file.writeAsBytesSync(bundle.toBuffer(), mode: io.FileMode.WRITE_ONLY);
    }
    // Process other libraries.
    library.importedLibraries
        .forEach((imported) => writeLibrarySummaries(seenLibs, imported));
    library.exportedLibraries
        .forEach((exported) => writeLibrarySummaries(seenLibs, exported));
  }
}

String _getLibraryHashString(String path) {
  String content =
      PhysicalResourceProvider.INSTANCE.getFile(path).readAsStringSync();
  List<int> hash = (new crypto.MD5()..add(content.codeUnits)).close();
  return crypto.CryptoUtils.bytesToHex(hash);
}

String _getSummaryPath(String path) {
  String baseLibraryName = pathos.basename(path);
  String hashHex = _getLibraryHashString(path);
  return '/Users/scheglov/tmp/summary_pkg/' + hashHex + '_' + baseLibraryName;
}

class MySummaryResultProvider extends ResultProvider {
  final InternalAnalysisContext context;
//  final SummaryTypeProvider typeProvider = new SummaryTypeProvider();

  MySummaryResynthesizer resynthesizer;
  SummaryResultProvider sdkHelper;

  MySummaryResultProvider(this.context) {
//    {
//      io.File file =
//          new io.File('/Users/scheglov/tmp/summary/sdk_analysis_summary');
//      if (file.existsSync()) {
//        _fillMaps(file);
//      }
//    }
//    print('unlinkedMap: $unlinkedMap');
//    print('prelinkedMap: $prelinkedMap');

    InternalAnalysisContext sdkContext = context.sourceFactory.dartSdk.context;
    context.typeProvider = sdkContext.typeProvider;
    sdkHelper = sdkContext.resultProvider;
    SummaryResynthesizer sdkResynthesizer = sdkHelper.resynthesizer;
    print('sdkResynthesizer: $sdkResynthesizer');

    resynthesizer = new MySummaryResynthesizer(sdkResynthesizer, context,
        context.typeProvider, context.sourceFactory, false);

//    if (resynthesizer == null) {
//      resynthesizer = new SummaryResynthesizer(null, context, typeProvider,
//          _getPrelinkedSummary, _getUnlinkedSummary, context.sourceFactory);
//      _buildCoreLibrary();
//      _buildAsyncLibrary();
//    }
  }

  bool compute(CacheEntry entry, ResultDescriptor result) {
    if (sdkHelper.compute(entry, result)) {
      return true;
    }
    AnalysisTarget target = entry.target;
//    if (result == TYPE_PROVIDER) {
//      entry.setValue(result, typeProvider, TargetedResult.EMPTY_LIST);
//      return true;
//    }
//    if (result == CONSTANT_VALUE) {
//      if (target.source != null && target.source.isInSystemLibrary) {
//        entry.setValue(result, target, TargetedResult.EMPTY_LIST);
//        return true;
//      }
//    }
    if (target is Source) {
      String uri = target.uri.toString();
//      if (!unlinkedMap.containsKey(uri) || !prelinkedMap.containsKey(uri)) {
//        return false;
//      }
      if (target.fullName
          .startsWith('/Users/scheglov/Source/Dart/sdk.git/sdk/pkg/analyzer')) {
        return false;
      }
//      print('MySummaryResultProvider: $result of $target');
      if (result == LIBRARY_ELEMENT1 ||
          result == LIBRARY_ELEMENT2 ||
          result == LIBRARY_ELEMENT3 ||
          result == LIBRARY_ELEMENT4 ||
          result == LIBRARY_ELEMENT5 ||
          result == LIBRARY_ELEMENT6 ||
          result == LIBRARY_ELEMENT7 ||
          result == LIBRARY_ELEMENT8 ||
          result == LIBRARY_ELEMENT ||
          false) {
        LibraryElement libraryElement = resynthesizer.getLibraryElement(uri);
        entry.setValue(result, libraryElement, TargetedResult.EMPTY_LIST);
        return true;
      } else if (result == READY_LIBRARY_ELEMENT2 ||
          result == READY_LIBRARY_ELEMENT5 ||
          result == READY_LIBRARY_ELEMENT6) {
        entry.setValue(result, true, TargetedResult.EMPTY_LIST);
        return true;
      } else if (result == SOURCE_KIND) {
        if (resynthesizer.linkedMap.containsKey(uri)) {
          entry.setValue(result, SourceKind.LIBRARY, TargetedResult.EMPTY_LIST);
          return true;
        }
        if (resynthesizer.unlinkedMap.containsKey(uri)) {
          entry.setValue(result, SourceKind.PART, TargetedResult.EMPTY_LIST);
          return true;
        }
        return false;
      }
    }
    return false;
  }
}

class MySummaryResynthesizer extends SummaryResynthesizer {
  final Map<String, UnlinkedUnit> unlinkedMap = <String, UnlinkedUnit>{};
  final Map<String, LinkedLibrary> linkedMap = <String, LinkedLibrary>{};

  MySummaryResynthesizer(SummaryResynthesizer parent, AnalysisContext context,
      TypeProvider typeProvider, SourceFactory sourceFactory, bool strongMode)
      : super(parent, context, typeProvider, sourceFactory, strongMode) {
    new io.Directory('/Users/scheglov/tmp/summary_pkg/')
        .listSync()
        .forEach((entry) {
      if (entry is io.File) {
        _fillMaps(entry);
      }
    });
  }

  @override
  LinkedLibrary getLinkedSummary(String uri) {
    return linkedMap[uri];
  }

  @override
  UnlinkedUnit getUnlinkedSummary(String uri) {
    return unlinkedMap[uri];
  }

  @override
  bool hasLibrarySummary(String uri) {
    return linkedMap.containsKey(uri);
  }

  void _fillMaps(io.File file) {
    List<int> buffer = file.readAsBytesSync();
    PackageBundle bundle = new PackageBundle.fromBuffer(buffer);
    for (int i = 0; i < bundle.unlinkedUnitUris.length; i++) {
      unlinkedMap[bundle.unlinkedUnitUris[i]] = bundle.unlinkedUnits[i];
    }
    for (int i = 0; i < bundle.linkedLibraryUris.length; i++) {
      linkedMap[bundle.linkedLibraryUris[i]] = bundle.linkedLibraries[i];
    }
  }
}

//class _NoSummaryException {
//  final String kind;
//  final String uri;
//
//  _NoSummaryException(this.kind, this.uri);
//
//  @override
//  String toString() => '_NoSummaryException($kind, $uri)';
//}
